#include <mutex>

#include "TROOT.h"
#include "TSystem.h"
#include "TFile.h"
#include "TH1F.h"
#include "TF1.h"
#include "TString.h"
#include "TRandom3.h"
#include "TPluginManager.h"
#include "Fit/Fitter.h"
#include "TMinuitMinimizer.h"

///////////////////////////////////////////////////////////////////////////////
//                                  get_hist                                 //
///////////////////////////////////////////////////////////////////////////////

class get_hist {
  
 public:
  
  get_hist(TString file_name) {
    m_hfile = new TFile(file_name,"read");
  }

  TH1F * get_test_hist() {
    if (m_hfile->GetListOfKeys()->Contains("histo")) {
      return (TH1F*) m_hfile->Get("histo");
    } else {
      return NULL;
    }
  }
    
  ~get_hist() {
    m_hfile->Close();
  }
  
 private:
  
  TFile *m_hfile;
};

///////////////////////////////////////////////////////////////////////////////
//                                  fit_hist                                 //
///////////////////////////////////////////////////////////////////////////////

std::mutex mtx;           // mutex for critical section

class fit_hist {

 public:
    
  fit_hist(TString file_name) : m_get_hist(file_name) {}

  void fit_test_hist(void) {


    ROOT::Math::MinimizerOptions::SetDefaultMinimizer("Minuit2");
    // We will be using Minuit2 for fitting since it is thread safe, so check it is there
    if (gROOT->GetPluginManager()->FindHandler("ROOT::Math::Minimizer", "Minuit2") == 0)  {
      std::cout<<"We need Minuit2 for parallel fitting!"<<std::endl;
      return;
    }
    TH1F * test_hist = m_get_hist.get_test_hist();
    // When fitting simple functions, such as a Gaussian, the initial
    // values of the parameters are automatically computed by ROOT.
    TF1* gaussian = new TF1("gaussian", "gaus", -3, 3);
    test_hist->Fit(gaussian);
    std::cout << "Peak = " << gaussian->GetParameter(0) << "\n"; // peak_fit
    std::cout << "Mean = " << gaussian->GetParameter(1) << "\n"; // mean_fit
    std::cout << "Sigma = " << gaussian->GetParameter(2) << "\n"; // sigma_fit
    delete gaussian;
  }

 private:

  get_hist m_get_hist;
};

///////////////////////////////////////////////////////////////////////////////
//                                   ctypes                                  //
///////////////////////////////////////////////////////////////////////////////

extern "C" {
  void analyze_file (const char * x_file_name);
  void create_files ();
  void enable_thread_safety();
}

void enable_thread_safety() {
  ROOT::EnableThreadSafety();
}

void analyze_file (const char * x_file_name) {
  mtx.lock();
  TString file_name(x_file_name);
  fit_hist fit(file_name);
  fit.fit_test_hist();
  mtx.unlock();
  
}
  
void create_files () {
  for (Int_t i = 1; i <= 2; ++i) {
    TString hfile_name;
    hfile_name.Form("histo%d.root", i);
    TFile hfile(hfile_name, "RECREATE", "Demo ROOT file for thread 1");
    TH1F *histo   = new TH1F("histo", "test histogram", 100, -3, 3);
    TRandom3 rndm;
    for (Int_t j = 0; j < 25000; ++j) {
      histo->Fill(rndm.Gaus(0,1)+0.1*(i-1));
    }
    hfile.Write();
    hfile.Close();
  }
}

